#ifndef TEMPSET_H
#define TEMPSET_H

#include <map>
#include <algorithm>
using namespace std;

template<class T> class TempSet;

/// Iteratore per TempSet
/**
 * L'iteratore  descritto da un puntatore al TempSet a cui si riferisce e da un iteratore
 * ad una multimap<T,int> che  contenuta nel TempSet.<br>
 * Sono realizzati i metodi minimi che ci servono: l'incremento, il confronto, la deferenziazione.
 */
template<class T> class TSiter{
	private:
		typedef multimap<T,int>::iterator ait;	///< tipo dell'iteratore ausiliario
		typedef const T& reference_type;	///< tipo reference
		TempSet<T>* cont;	///< container corrente
		ait curr;	///< itaratore ausiliario corrente

	public:
		/// costruttore banale
		TSiter() {};
		/// costruttore utile
		/**
		 * \param x il TempSet di riferimento
		 * \param v un iteratore alla multimap del TempSet di riferimento
		 */
		TSiter(TempSet<T>& x, ait v): cont(&x), curr(v) {};
		/// incremento (serve per scandire un intervallo)
		TSiter operator++();
		/// uguaglianza
		template<class C> friend bool operator==(const TSiter<C>& sx, const TSiter<C>& dx);
		/// deferenziazione
		reference_type operator*() const;
};

/// Multiset temporeneo
/**
 * La classe TempSet fornisce una versione temporanea del multiset.<br>
 * Insieme all'oggetto da inserire viene passata un'eventuale data di scadenza.<br>
 * Il TempSet si occuper di ignorare gli oggetti scaduti e di rimuoverli
 * di tanto in tanto dalla memoria.<br>
 * La classe  formata da una multimap<T,int> usata per memorizzare le coppie (oggetto,scadenza).
 */
template<class T> class TempSet{
	private:
		/// l'incremento dell'iteratore deve essere friend
		friend TSiter<T> TSiter<T>::operator++();
		typedef TSiter<T> iterator;	///<tipo iteratore
		typedef multimap<T,int>::iterator ait;	///< tipo iteratore ausiliario
		/// multimap per memorizzare i dati
		multimap<T,int> TS;
		/// contatore di find e equal_range
		int count;
		/// numero di find da far passare prima di una pulizia
		static const int clean=50;
		/// toglie gli oggetti scaduti dalla memoria
		void purge();
	public:
		/// iteratore all'inizio del TempSet
		iterator begin();
		/// iteratore a dopo la fine del TempSet
		iterator end();
		/// cerca un oggetto
		/**
		 * \return l'iteratore all'oggetto trovato o end() se non trovato
		 */
		iterator find(const T&);
		/// cerca un intervallo di oggetti
		/**
		 * \return una coppia di iteratori al primo e subito dopo
		 * l'ultimo degli oggetti trovati. entrambi end() in caso di ricerca fallita
		 */
		pair<iterator, iterator> equal_range(const T&);
		/// inserisce un oggetto con eventuale scadenza
		void insert(const T&, int=0);
};

/// predicato controllo scadenza
/**
 * usato da TempSet<T>::purge() per controllare quali dati
 * rimuovere perch scaduti
 */
template<class C> class old : public unary_function< pair<const C,int>, bool>{
	public:
		/// controlla scadenza
		bool operator()(pair<const C,int> cop) {
			return cop.second<100;
		}
};

template<class T> TSiter<T>::reference_type TSiter<T>::operator*() const{
	// il valore  memorizzato come primo elemento della coppia corrente
	return curr->first;
}

template<class T> TSiter<T> TSiter<T>::operator++(){
	// incrementa punatore ausiliario
	ait it=++curr;
	// incrementalo ancora finch non punta ad un elemento non vecchio
	// o alla fine del container
	while(it!=cont->TS.end() && old<T>()(*it))
		++it;
	curr=it;
	
}

template<class T> bool operator==(const TSiter<T>& sx, const TSiter<T>& dx){
	// due TSiter sono uguali se si riferiscono allo stesso
	// container con lo stesso iteratore ausiliario
	return sx.cont==dx.cont && sx.curr==dx.curr;
}

template<class T> TempSet<T>::iterator TempSet<T>::begin(){
	// prendi il primo elemento come iteratore ausiliario
	ait b=TS.begin();
	// incrementalo finch non punta ad un elemento non vecchio
	// o alla fine del container
	while (b!=TS.end() && old<T>()(*b))
		++b;
	return TSiter<T>(*this, b);
}

template<class T> TempSet<T>::iterator TempSet<T>::end(){
	// restituisce l'iteratore alla fine del container
	return TSiter<T>(*this, TS.end());
}

template<class T> TempSet<T>::iterator TempSet<T>::find(const T& what){
	// se sono passati clean find() dall'ultima spurgata spurga
	if (++count%clean==0)
		purge();
	// cerca gli elementi cirrispondenti nella map
	pair<ait,ait> rang=TS.equal_range(what);
	// prendi il primo non vecchio
	ait ret=rang.first;
	while (ret!=rang.second &&  old<T>()(*ret))
		++ret;
	// se esiste restituiscilo, altrimenti restituisci end()
	return ret==rang.second? TSiter<T>(*this, TS.end()): TSiter<T>(*this, ret);
}

template<class T> void TempSet<T>::purge(){
	// crea una copia di TS senza gli elementi vecchi
	multimap< T,int > purged;
	remove_copy_if(TS.begin(), TS.end(), inserter(purged,purged.end()), old<T>() );
	// sostituisci a TS la nuova map ripulita
	TS=purged;		 
}

template<class T> pair<TempSet<T>::iterator,TempSet<T>::iterator> TempSet<T>::equal_range(const T& what){
	// se sporca da un po' pulisci la map
	if (++count%clean==0)
		purge();

	// cerca gli elementi corrispondenti
	pair<ait,ait> rang=TS.equal_range(what);

	// cerca il primo elemento non vecchio
	ait f=rang.first;
	while (f!=rang.second && old<T>()(*f))
		++f;
	
	iterator fi,la;
	// se non ci sono elementi non vecchi
	if (f==rang.second){
		// restituisci end() come range
		fi=la=TSiter<T>(*this, TS.end());
	// altrimenti
	} else {
		// restituisci il primo buono come primo
		fi=TSiter<T>(*this, f);
		// e per l'ultimo controlla che non sia vecchio
		// e sostituiscilo col prox buono
		ait l=rang.second;
		while (l!=TS.end() && old<T>()(*l))
			++l;
		la=TSiter<T>(*this, l);
	}

	// restituisci il range calcolato
	return make_pair(fi,la);
}

template<class T> void TempSet<T>::insert(const T& what, int time=0){
	// inserisci una coppia
	// essendo TS una multimap non crea problemi di duplicazione di elementi
	// e possiamo lasciare insert come void
	TS.insert(make_pair(what,time));
}

#endif
